#!/usr/bin/python
#
# Copyright 2018-2021 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List


class BaseManager:
    CONFIG = None

    def __init__(self, config):
        if not isinstance(config, self.CONFIG):
            raise ValueError(
                "The current search manager `{}` is not compatible "
                "with the search kind `{}` defined in the config.".format(
                    self.CONFIG.IDENTIFIER, config.IDENTIFIER
                )
            )
        self.config = config

    def get_suggestions(self, **kwargs) -> List[Dict]:
        raise NotImplementedError
